/*
 Copyright 2002, 2003 Alexandre Gil
                                                                                
 This file is part of mwu-extractor.
                                                                                
 mwu-extractor is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
                                                                                
 mwu-extractor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
                                                                                
 You should have received a copy of the GNU General Public License
 along with mwu-extractor; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nome: Vector.cpp
// Objectivo: Implementacao da classe generica Vector.
//            Inspirada no livro do Prof. Pedro Guerreiro (UNL/FCT) mas modificado.
// Autor: Alexandre Gil
// Data: 23 Abr 2002

#include <iostream>
#include <iomanip>

#include "Vector.h"
#include "MyUtils.h"

template <class T>
Vector<T>::Vector(unsigned long capacity):
  items(new T[capacity]),
  capacity(capacity),
  filled(0),
  growFactor(1.50),
  debug(false)
{
}

template <class T>
Vector<T>::Vector(const Vector<T>& other):
  items(new T[other.capacity]),
  capacity(other.capacity),
  filled(0),
  growFactor(other.growFactor),
  debug(false)
{
	for(unsigned long i = 0; i < capacity; i++)
		items[i] = other.items[i];
}

template <class T>
Vector<T>::~Vector()
{
	if(capacity > 0) delete [] items;
}

template <class T>
void Vector<T>::Copy(const Vector<T>& other)
{
	delete [] items;
	items = new T[other.capacity];
	capacity = other.capacity;
	growFactor = other.growFactor;
	filled = other.filled;
	for(unsigned long i = 0; i < capacity; i++)
		items[i] = other.items[i];
}

template <class T>
Vector<T>& Vector<T>::operator=(const Vector<T>& other)
{
	Copy(other);
	return *this;
}

template <class T>
unsigned long Vector<T>::Capacity() const
{
	return capacity;
}

template <class T>
unsigned long Vector<T>::Filled() const
{
	return filled;
}

template <class T>
void Vector<T>::CleanLast()
{
	filled--;
}

template <class T>
float Vector<T>::GrowFactor() const
{
	return growFactor;
}

template <class T>
void Vector<T>::SetGrowFactor(float newGrowFactor)
{
	growFactor = newGrowFactor;
}

template <class T>
void Vector<T>::Grow()
{
	GrowTo(static_cast<unsigned long>(capacity*growFactor));
}

template <class T>
void Vector<T>::GrowTo(unsigned long newCapacity)
{
	if(newCapacity > capacity)
	{
		T * temp = items;
		items = new T[newCapacity];
		for(unsigned long i = 0; i<capacity;i++)
			items[i] = temp[i];
		capacity = newCapacity;
		delete [] temp;
	}
}

template<class T>
bool Vector<T>::Has(const T& s) const
{
	for(unsigned long i=0; i < filled; i++)
	  if(items[i] == s)
	    return true;
	return false;	
}

template<class T>
unsigned long Vector<T>::CountIf(const T& s) const
{
	unsigned long result = 0;
	for(unsigned long i = 0; i < filled;i++)
		if(items[i] == s)
			result++;
	return result;
}

// Retorna primeiro indice dum elemento com valor s
template<class T>
unsigned long Vector<T>::IndexOf(const T& s) const    //devolve Filled() se nao encontrar
{
	unsigned long i;

	for(i=0; i < filled; i++)
	  if(items[i] == s)
			break;
	return i;
}

// Retorna primeiro indice dum elemento com valor s utilizando pesquisa binaria
template<class T>
unsigned long Vector<T>::IndexOfBin(const T& s) const    //devolve Filled() se nao encontrar
{
	unsigned long i = 0;
	unsigned long j = filled-1;
	unsigned long med;

	while(j >= i)
	{
		med = (j+i) / 2;
	 	if(items[med] == s)
			return med;
		if(items[med] > s)
		{
			if(med == 0)
				break;
		  j = med-1;
		}
		else
		{
			if(med == filled -1)
				break;
			i = med+1;
		}
	}
	return filled;
}

// Retorna ltimo indice dum elemento com valor s
template<class T>
unsigned long Vector<T>::LastIndexOf(const T& s) const
{
	unsigned long i;
	
	for(i=filled-1; i >= 0; i--)
	  if(items[i] == s)
	  	break;
  return i;
}

template<class T>
T& Vector<T>::At(unsigned long x) const
{
	return items[x];
}

template<class T>
void Vector<T>::PutAt(unsigned long x, const T& s)
{
	if(x == capacity)
	  Grow();
	items[x] = s;
	if(x >= filled) filled = x+1;
}

template<class T>
void Vector<T>::Add(const T& s)
{
	if(filled == capacity)
	  Grow();
	items[filled++] = s;
}


template <class T>
void Vector<T>::Swap(unsigned long x, unsigned long y)
{
	T m = items[x];
	items[x] = items[y];
	items[y] = m;
}

template <class T>
void Vector<T>::Write(std::ostream& output) const
{
	output << "Imprimir Vector " << std::endl;
	output << "--------------- " << std::endl;
	for(unsigned long i = 0; i < filled; i++)
	  output << std::setw(6) << std::setfill('0') << i << " - " << items[i] << std::endl;;
	output << "--------------- " << std::endl;
}

//Ternary-Split Quicksort baseado no algoritmo do Bentley-McIlroy (1993) descrito em
//"Fast Algorithms for Sorting and Searching Strings" de Bentley-Sedgewick
// Versao do algoritmo onde os elementos do vector sao strings
template<class T>
void Vector<T>::TQuickSort(unsigned long l, unsigned long r)
{
	if((r-l)+1 == 2)
	{
		if(items[l] > items[r])
			Swap(l,r);
	}
	else
	{
	  unsigned long k, s, t, w;

		bool jNegativo = false;

		unsigned long pivot = ChoosePivot(l,r);
	  T v = items[pivot];

	  Swap(l,pivot);

		unsigned long i = l+1;
		unsigned long p = l+1;
	  unsigned long j = r;
		unsigned long q = r;

	  for(;;)
	  {
	    while(i <= j && items[i] <= v)
	    {
			  if(items[i] == v)
			  {
			  	Swap(p, i);
			  	p++;
			  }
	    	i++;
	    }
	    while(!jNegativo && j >= i && items[j] >= v)
	    {
				if(items[j] == v)
				{
					Swap(q, j);
					q--;
				}	
	  	if(j==0)
	  		jNegativo = true;
	  	else
				j--;
	    }
	    if (i > j || jNegativo)
	    	break;
	  	Swap(i, j);
	  	i++;
	  	if(j==0)
	  		jNegativo = true;
	  	else
				j--;
	  }
		s = MIN(p-l,i-p);
	  for(k = l, w = i-s; s ; s--, k++, w++)    // move equals from left to center
	  	Swap(k, w);

		s = (jNegativo? MIN(q+1,r - q) : MIN(q - j,r - q));
	  for(k = i, w = r-s+1; s; s--, k++, w++)    // mode equals from right to center
	  	Swap(k, w);

		s=i-p;
		if(s>1)
			TQuickSort(l, l+s-1);

		if((p-l+r-q) > 1)
			TQuickSort(i-(p-l), i+r-q-1);

		t= (jNegativo? q+1: q-j);
		if(t>1)
	  	TQuickSort(r-t+1, r);
	}
}

// Subrotina para escolha do pivot para o algoritmo SplitSort de Bentley-McIlroy
template<class T>
unsigned long Vector<T>::ChoosePivot(unsigned long l, unsigned long r) const
{
	unsigned long nr = r - l + 1;       // numero de items
	unsigned long pm = l + nr/2;				// ponto medio
	if(nr>7)
	{
		unsigned long pl = l;
		unsigned long pn = r;
		if(nr>40)
		{
			unsigned long s = nr >> 3;
			pl = VMED3(pl, pl+s, pl+s+s);
			pm = VMED3(pm-s, pm, pm+s);
			pn = VMED3(pn-s-s, pn-s, pn);		
		}
		pm=VMED3(pl, pm, pn);
	}
	return pm;
}

template<class T>
void Vector<T>::Sort(unsigned long l, unsigned long r)
{
	TQuickSort(l,r);
}


template <class T>
void Vector<T>::Reset()
{
	filled = 0;
}

template <class T>
void Vector<T>::SetDebug(bool state)
{
	debug = state;
}
